#!/usr/bin/perl -w
###############################################################################
#
# WinMac.pl - mactimes sorter. Part of the DirTools project.
#
# $Revision: 1.3 $
#
# Copyright (c) 2001-2004 Olivier Castan
#
# This program/include file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program/include file is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (in the main directory of the Linux-NTFS
# distribution in the file COPYING); if not, write to the Free Software
# Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
###############################################################################

=pod

WinMac.pl

=head1 SYNOPSIS

B<WinMac.pl>
[B<-h>]
[B<-D>]
[B<-i> mactime file]
[B<-u> users file]
[B<-c> fields list]
[B<-t> timestamps list]
[B<-f> tab|latex|html]
[B<-o> output file]
start_date[-end_date]

=head1 DESCRIPTION

=head1 OPTIONS

=over 4

=item B<-h>

Display options help.

=item B<-d>

Debug, more informations printed on stderr.

=item B<-i>

mactime file to sort (created with dirmft(.exe) or dirhive(.exe)). If not
present will try to read from STDIN.

=item B<-c>

Comma separated list of fields to output in result table (default 
depends of the source). Supported fields are: 'MFT Index', 'FileName', 
'# Links', 'Size', 'Attributes', '# Data Streams', 'Owner', 'Group',
'KeyNode', '# Subkeys', '# Values' and 'mactime' (for the sort timestamp).

=item B<-t>

Comma separated list of timestamps fields used to sort (default 
is to use all). Supported fields are: 'Last Write', 'Creation', 
'Attribute Modification' and 'Last Access'.

=item B<-u>

user file to convert SID to full names (created with dirmft.exe 
dirhive.exe - Windows versions only).

=item B<-f>

Output format : tab separated fields, html or latex tables 
(default is html).

=item B<-o>

Output to file instead of stdout.

=back

=head1 EXAMPLES

=head1 AUTHOR

Olivier Castan E<lt>F<olivier.castan@certa.ssi.gouv.fr>E<gt>

=cut

####################################################################
# Intialisation
####################################################################
use strict;
use Getopt::Std;
use Pod::Usage;
use Date::Manip;

use vars qw($opt_h $opt_D $opt_i $opt_o $opt_f $opt_u $opt_c $opt_t);

my $date_courante = &texte_date(localtime time);

# Month output
#my @chiffre_mois = ("janv", "fvr", "mars", "avri", "mai ",
#                 "juin", "juil", "aout", "sept", "octo",
#                 "nove", "dce");
my @chiffre_mois = ("jan", "feb", "mar", "apr", "may",
          "jun", "jul", "aug", "sep", "oct",
          "nov", "dec");

my %date_letter = ('Last Write' => 'w',
          'Creation' => 'c',
          'Attribute Modification' => 'm',
          'Last Access' => 'a');

#                    Field         Title   Function Args  Format
my %champs_config = ('MFT Index'=>
               ['Index', undef, undef, {'html'=>'CENTER', 'latex'=>'c'}],
          'FileName'=>
               ['File', \&check_efface, ['# Links'],
               {'html'=>'name', 'latex'=>'p{10cm}'}],
          '# Links'=>['# links', undef, undef, {'html'=>'num', 'latex'=>'r'}],
          'Size'=>['Size', undef, undef, {'html'=>'num', 'latex'=>'r'}],
          'Attributes'=>
               ['Attributes', undef, undef, {'html'=>'other', 'latex'=>'c'}],
          '# Data Streams'=>
               ['# streams', undef, undef, {'html'=>'num', 'latex'=>'r'}],
          'Owner'=>
               ['Owner', \&sid2nom, undef, {'html'=>'sid', 'latex'=>'c'}],
          'Group'=>
               ['Group', \&sid2nom, undef, {'html'=>'sid', 'latex'=>'c'}],
          'KeyNode'=>
               ['KeyNode', undef, undef, {'html'=>'name', 'latex'=>'p{10cm}'}],
          '# Subkeys'=>
               ['# subkeys', undef, undef, {'html'=>'num', 'latex'=>'r'}],
          '# Values'=>
               ['# values', undef, undef, {'html'=>'num', 'latex'=>'r'}],
          'mactime'=>[undef, undef, undef, {'html'=>'mac', 'latex'=>'c'}]);

my %out_defaults = ('dirmft'=>{ 'champs'=>['Size', 'mactime', 'Attributes',
                         'MFT Index', 'Owner', 'Group', 'FileName'],
                    'dates'=>['Attribute Modification', 'Last Access',
                         'Creation', 'Last Write'],
                    'nom'=>'FileName',
                    'titre'=>'files' },
                    'dirhive'=>{ 'champs'=>['# Values', 'Owner', 'Group',
                         'KeyNode'],
                    'dates'=>['Last Write'],
                    'nom'=>'KeyNode',
                    'titre'=>'keys' });

my @out_champs = ();

my @out_dates = ();

my $sorted_type = 'values';

my @out_style = ();

my $tri2_idx = undef;

my @mac_affiche = ();

my %liste_user = ('S-1-5-32-544', 'BUILTIN\\ADMINISTRATORS',
          'S-1-5-32-545', 'BUILTIN\\USERS',
          'S-1-5-32-546', 'BUILTIN\\GUESTS',
          'S-1-5-32-547', 'BUILTIN\\POWER USERS',
          'S-1-5-32-548', 'BUILTIN\\ACCOUNT OPERATORS',
          'S-1-5-32-549', 'BUILTIN\\SERVER OPERATORS',
          'S-1-5-32-550', 'BUILTIN\\PRINT OPERATORS',
          'S-1-5-32-551', 'BUILTIN\\BACKUP OPERATORS',
          'S-1-5-32-552', 'BUILTIN\\REPLICATOR',
          'S-1-5-32-553', 'BUILTIN\\RAS SERVERS',
          'S-1-5-32-554', 'BUILTIN\\PRE-WINDOWS 2K COMP. ACCESS',
          'S-1-5-32-555', 'BUILTIN\\REMOTE DESKTOP USERS',
          'S-1-5-32-556', 'BUILTIN\\NETWORK CONF. OPERATORS',
          'S-1-0-0', '\\NULL',
          'S-1-1-0', '\\EVERYONE',
          'S-1-2-0', '\\LOCAL',
          'S-1-3-0', '\\CREATOR OWNER',
          'S-1-3-1', '\\CREATOR GROUP',
          'S-1-3-2', '\\CREATOR OWNER SERVER',
          'S-1-3-3', '\\CREATOR GROUP SERVER',
          'S-1-5-1', 'NT AUTHORITY\\DIALUP',
          'S-1-5-2', 'NT AUTHORITY\\NETWORK',
          'S-1-5-3', 'NT AUTHORITY\\BATCH',
          'S-1-5-4', 'NT AUTHORITY\\INTERACTIVE',
          'S-1-5-6', 'NT AUTHORITY\\SERVICE',
          'S-1-5-7', 'NT AUTHORITY\\ANONYMOUS LOGON',
          'S-1-5-8', 'NT AUTHORITY\\PROXY',
          'S-1-5-9', 'NT AUTHORITY\\SERVER LOGON',
          'S-1-5-10', 'NT AUTHORITY\\SELF',
          'S-1-5-11', 'NT AUTHORITY\\AUTHENTICATED USER',
          'S-1-5-12', 'NT AUTHORITY\\RESTRICTED CODE',
          'S-1-5-13', 'NT AUTHORITY\\TERMINAL SERVER',
          'S-1-5-14', 'NT AUTHORITY\\REMOTE INTERACTIVE LOGON',
          'S-1-5-18', 'NT AUTHORITY\\SYSTEM',
          'S-1-5-19', 'NT AUTHORITY\\LOCALSERVICE',
          'S-1-5-20', 'NT AUTHORITY\\NETWORKSERVICE');

my %known_users = ('500', 'ADMINISTRATOR',
             '501', 'GUEST',
             '502', 'KRBTGT',
             '512', 'DOMAIN ADMINS',
             '513', 'DOMAIN USERS',
             '514', 'DOMAIN GUESTS',
             '515', 'DOMAIN COMPUTERS',
             '516', 'DOMAIN CONTROLLERS',
             '517', 'CERT PUBLISHERS',
             '518', 'SCHEMA ADMINS',
             '519', 'ENTERPRISE ADMINS',
             '520', 'GROUP POLICY CREATOR OWNERS',
             '533', 'RAS AND IAS SERVERS');

my $domain_count = 1;

my %domain_list = ('32'=>'BUILTIN');

# Program name
$0 = rindex($0, "/") > -1 ? substr($0, rindex($0, "/")+1) : $0;

####################################################################
# Analysing command line and options handling
####################################################################

&getopts("hDc:i:u:f:o:") ||
     die pod2usage({-verbose=>0, -exitval=>2});

# Help
if ($opt_h) {
     pod2usage({-exitval=>1,
          -verbose=>1,
          -message=>"More informations with 'perldoc $0'\n"});
     exit;
}

# Input file
if ($opt_i) {
     print STDERR "Mactime file: $opt_i\n" if $opt_D;
} else {
     print STDERR "Reading mactimes from STDIN\n" if $opt_D;
}

# Output format
if ($opt_f) {
     $opt_f = lc $opt_f;
     die pod2usage({-verbose=>0, -exitval=>2})
          unless $opt_f =~ /^(tab|html|latex)$/;
} else {
     $opt_f = 'html';
}

# Output file
if ($opt_o) {
     if (-e $opt_o) {
          # I/O may be redirected
          # directly open the tty
          if(open(HTTY, "+</dev/tty")) {
               select((select(HTTY), $| = 1)[0]); # non buffered I/O
               print HTTY "Warning $opt_o may be overwritten...[Y/n]";
               my $reponse = getc HTTY;
               print HTTY "\n";
               close HTTY;
               exit if $reponse =~ /^n/;
          }
     }
     open(OHANDLE, ">$opt_o") ||
          die "Can't write to $opt_o : $!\n";
     print STDERR "Output file : $opt_o\n" if $opt_D;
     select OHANDLE;
}

# Read user file
if ($opt_u) {
     my $user = &lecture_table( $opt_u );
     my $SID_index = $user->{'champs'}{'SID'};
     my $detail_index = $user->{'champs'}{'Name'};

     for my $index (0..$#{$user->{'donnes'}}) {
          my $hash_ref = \@{$user->{'donnes'}[$index]};
          my $SID = $hash_ref->[$SID_index];
          my $detail = $hash_ref->[$detail_index];

          print STDERR "$SID = $detail\n" if $opt_D;
          if ($SID) {
               if ($detail) {
                    $liste_user{$SID} = $detail;
               } elsif (! exists($liste_user{$SID})) {
                    $liste_user{$SID} = &sid2nom($SID);
               }
          }
     }
}

# Fields to output
if ($opt_c) {
     @{$out_champs[0]} = split /,/, $opt_c;
     print STDERR 'User specified fields '.join("\t", @{$out_champs[0]}) .
          "\n" if $opt_D;
}

# Timestamps to use
if ($opt_t) {
     @{$out_dates[0]} = split /,/, $opt_t;
     print STDERR 'User specified timestamps '.join("\t", @{$out_dates[0]}) .
          "\n" if $opt_D;
}

# Reading date(s) argument(s)
die pod2usage({-verbose=>0, -exitval=>2}) if ($#ARGV != 0);
my ($temps_init, $temps_fin) = ($ARGV[0], undef);
if ($temps_init =~ /-/) {
     ($temps_init, $temps_fin) = split(/-/, $temps_init);
}
#&Date_Init("Language=French","DateFormat=non-US", "ConvTZ=IGNORE");
&Date_Init("DateFormat=non-US", "ConvTZ=IGNORE");
print STDERR "Start date: $temps_init\n" if $opt_D;
my $parse_temps1 = &ParseDate($temps_init);
die "Wrong date format: $temps_init\n" unless ($parse_temps1);
my $seconde_inf = &UnixDate($parse_temps1, "%q");
my $seconde_sup = 0;
if ($temps_fin) {
     print STDERR "End date: $temps_fin\n" if $opt_D;
     my $parse_temps2 = &ParseDate($temps_fin);
     die "Wrong date format: $temps_fin\n" unless ($parse_temps2);
     $seconde_sup = &UnixDate($parse_temps2, "%q");
}

####################################################################
# Input file analysis
####################################################################
# Read main input file
my $resultat;
if ($opt_i) {
     $resultat = &lecture_table($opt_i);
} else {
     $resultat = &lecture_table(undef);
     $opt_i = "(STDIN)";
}

my $type_fichier = '';

if (($type_fichier = $resultat->{'origine'}{'Content'}) &&
     exists $out_defaults{$type_fichier}) {
     $out_champs[0] = $out_defaults{$type_fichier}{'champs'} unless $opt_c;
     $out_dates[0] = $out_defaults{$type_fichier}{'dates'} unless $opt_t;
     $sorted_type = $out_defaults{$type_fichier}{'titre'};
     print STDERR "File type: $type_fichier\n" if $opt_D;
} else {
     my $message = "Can't find file type or unknown type:\n";
     map{$message .= "'$_'='$resultat->{'origine'}{$_}'\n"}
          keys %{$resultat->{'origine'}};
     die $message;
}

# Identify fields to show and their offsets
my $mactime_found = 0;
foreach my $index (@{$out_champs[0]}) {
     if ($index eq 'mactime') {
          push @{$out_champs[1]}, undef;
          $mactime_found = 1;
          print STDERR "Special field: 'mactime'\n" if $opt_D;
     } else {
          die "Field '$index' not found in $opt_i\n"
               unless exists $resultat->{'champs'}{$index};
          push @{$out_champs[1]}, $resultat->{'champs'}{$index};
          print STDERR "Field: '$index'\t$resultat->{'champs'}{$index}\n"
               if $opt_D;
     }
     push @{$out_champs[2]}, $champs_config{$index}[1];
     map {$_ = $resultat->{'champs'}{$_}} @{$champs_config{$index}[2]}
          if defined $champs_config{$index}[2];
     push @{$out_champs[3]}, $champs_config{$index}[2];
}

#  Identify timestamp fields to use and their offsets
my $valeur = 1;
foreach my $index (@{$out_dates[0]}) {
     die "Timestamp field '$index' not found in $opt_i\n"
          unless exists $resultat->{'champs'}{$index};
     push @{$out_dates[1]}, $resultat->{'champs'}{$index};
     push @{$out_dates[2]}, $valeur;
     print STDERR "Timestamp field: '$index'\t$resultat->{'champs'}{$index}".
          "\t'$valeur'\n" if $opt_D;
     $valeur *= 2;
}

#  Table to display mactime field
if ($mactime_found) {
     for (my $index = 0; $index < $valeur; $index++) {
          my $chaine = '';
          for my $index2 (0..$#{$out_dates[0]}) {
               $chaine .= ($index & $out_dates[2]->[$index2]?
                    $date_letter{$out_dates[0]->[$index2]} :'.');
          }
          push @mac_affiche, $chaine;
          print STDERR "Encoding: $chaine\t$index\n" if $opt_D;
     }
}

die "Field $out_defaults{$type_fichier}{'nom'} not found in $opt_i"
     unless exists $out_defaults{$type_fichier}{'nom'};
$tri2_idx = $resultat->{'champs'}{$out_defaults{$type_fichier}{'nom'}};
print STDERR "Secondary sort key: $out_defaults{$type_fichier}{'nom'}".
     "\t$resultat->{'champs'}{$out_defaults{$type_fichier}{'nom'}}\n" if $opt_D;

# Data collection
my %out_liste = ();
foreach my $index (@{$resultat->{'donnes'}}) {
     my $append = $index->[$tri2_idx];
     for my $index2 (0..$#{$out_dates[1]}) {
          my $value = $index->[$out_dates[1]->[$index2]];
          my $date_val = &UnixDate( &ParseDate( $value), "%q");
          # "inodes" 16 to 24 (MFT extensions)
          # don't have timestamps when unused...
          next unless defined $date_val;
          if ( $date_val >= $seconde_inf && ( (! $seconde_sup) ||
               $date_val <= $seconde_sup)) {
               my $clef = "$date_val,". substr( $value, rindex($value, '.') +
                    1, 3) . ",$append";
               $out_liste{$clef}[0] |= $out_dates[2]->[$index2];
               $out_liste{$clef}[1] = $index;
               print STDERR "Store to '$clef' ". join('|', @$index).
                    ", criteria '$out_dates[0]->[$index2]'\n" if $opt_D;
          }
     }
          
}

# Result output
my $memo_date ='';
&print_debut;
for my $clef (sort keys %out_liste) {
     my @ligne = ();
     my ($date, $milli) = split(/,/, $clef);
     my ($annee, $mois, $jour_m, $heure, $min, $sec) =
          ($date =~ /(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)/);
     my $chaine_date = "$jour_m $chiffre_mois[$mois-1] $annee ".
          "$heure:$min:$sec.$milli";
     if ($chaine_date eq $memo_date) {
          $chaine_date ='';
     } else {
          $memo_date = $chaine_date;
     }
     push @ligne, $chaine_date;
     for my $index (0..$#{$out_champs[1]}) {
          my $pos = $out_champs[1]->[$index];
          my $ftn = $out_champs[2]->[$index];
          my $args = $out_champs[3]->[$index];
          if (defined $pos) {
               my $val = $out_liste{$clef}[1][$pos];
               my @ftn_args = ();
               @ftn_args = map{$out_liste{$clef}[1][$_]} @$args
                    if defined $args;
               $val = &$ftn($val, @ftn_args) if defined $ftn;
               push @ligne, $val;
          } else {
               push @ligne, $mac_affiche[$out_liste{$clef}[0]];
          }
     }
     &print_ligne_tableau(\@ligne);
}
&print_fin;

foreach my $clef (keys %domain_list) {
     print STDERR "$domain_list{$clef} : S-1-5-$clef\n";
}

close;
exit(0);

####################################################################
# Sub Routines
####################################################################

# Create a textual date with input from
# gmtime or localtime
sub texte_date {
    my ($seconde, $minute, $heure, $jmois, $mois, $annee, $jsem) = @_;    
#    use constant JOUR_FR => qw(dimanche lundi mardi mercredi jeudi
#            vendredi samedi);
#    use constant MOIS_FR => qw(janvier fvrier mars avril mai juin
#            juillet aot septembre octobre novembre dcembre);
#
#    return (JOUR_FR)[$jsem]." $jmois ".(MOIS_FR)[$mois].' '.($annee + 1900).
#            ' '."$heure:". sprintf("%02u:%02u", $minute, $seconde);
    use constant DAY_EN => qw(sunday monday tuesday wednesday thursday
            friday saturday);
    use constant MONTH_EN => qw(january february march april may june
            july august september october november december);

    return (DAY_EN)[$jsem]." $jmois ".(MONTH_EN)[$mois].' '.($annee + 1900).
            ' '."$heure:". sprintf("%02u:%02u", $minute, $seconde);
}

sub check_efface {
     my ($nom, $nb_liens) = @_;

     return $nom if $nb_liens;
     if ($opt_f eq 'html') {
          return "<SPAN CLASS='erased'>&lt;$nom&gt;</SPAN>";
     } else {
          return "<$nom>";
     }
}

sub sid2nom {
     my $sid = shift;

     return $liste_user{$sid} if exists $liste_user{$sid};
     return $sid unless $sid =~ /S-1-5-([-\d]+)-(\d+)/;
     $domain_list{$1} = "DOMAIN ".$domain_count++ if ! exists $domain_list{$1};
     $liste_user{$sid} = "$domain_list{$1}\\".($known_users{$2} || $2);
     return $liste_user{$sid};
}

sub print_debut {
     my @affiche_head = ('UTC Date/Time', @{$out_champs[0]});
     @out_style = @affiche_head;
     map {$_ = ($_ eq 'mactime'? uc $mac_affiche[-1]:($champs_config{$_}[0] ||
          $_))} @affiche_head;
     map {$_ = $champs_config{$_}[3]{$opt_f}} @out_style;
     if ($opt_f eq 'html') {
          print <<"HTMLHEADEOT";
<HTML>
<HEAD>
  <STYLE TYPE='text/css'>
    .erased {color: green;}
    TD.sid {text-align: center;}
    TD.num {text-align: right; color: blue;}
    TD.name {text-align: left;}
    TD.mac {text-align: center; color: red;}
    TD.other {text-align: center;}
  </STYLE>
  <TITLE>Timestamp sorted $sorted_type</TITLE>
</HEAD>
<BODY>
  <P>Analysis date: $date_courante<BR>
  <TABLE BORDER=1>
HTMLHEADEOT
          print '    <TH>';
          print join "</TH>\n    <TH>", @affiche_head;
          print "</TH>\n";
     } elsif ($opt_f eq 'latex') {
          $0 =~  s/([_&#])/\\$1/g;
          print <<"LATEXHEADEOT";
\\documentclass[a4paper]{article}
\\usepackage[T1]{fontenc}
\\usepackage[latin1]{inputenc}
\\usepackage{hyperref}
\\usepackage{longtable}
\\begin{document}
\\title{Timestamp sorted $sorted_type}
\\date{$date_courante}
\\author{generated by \\emph{$0}}
\\maketitle
Analysis date: $date_courante

\\begin{scriptsize}
\\begin{longtable}{|
LATEXHEADEOT
          map { print ((defined $_?$_:'c'),'|')} @out_style;
          print "}\n";
          print "\\hline\n";
          print join '&', @affiche_head;
          print "\\\\\n\\hline\\hline\n\\endhead\n";
     } else {
          print "TIMESTAMP SORTED ".uc($sorted_type)."\n\n";
          print "Analysis date : $date_courante\n\n";
          print join "\t", @affiche_head;
          print "\n";
     }
}
          
sub print_ligne_tableau {
     my $ref = shift;
     my $sortie;
     if ($opt_f eq 'html') {
          $sortie = "    <TR>\n";
          for my $index (0..$#{$ref}) {
               $sortie .= '      <TD' . (defined $out_style[$index]?
                    " CLASS=\"$out_style[$index]\"":'').
                    ">$ref->[$index]</TD>\n";
          }
          $sortie .= "    </TR>\n";
     } elsif($opt_f eq 'latex') {
          map {s/\\/\\char\`\\\\\{\}/g} @$ref;
          map {s/(?<!\$\\backslash)\$(?!\\backslash\$)/\\\$/g} @$ref;
          map {s/([_&#])/\\$1/g} @$ref;
          $sortie = join('&', @$ref) . "\\\\\n\\hline\n";
     } else {
          $sortie = join("\t", @$ref). "\n";
     }
     print $sortie;
}

sub print_fin {
     if ($opt_f eq 'html') {
          print "  </TABLE><BR>\n</BODY>\n</HTML>\n";
     } elsif ($opt_f eq 'latex') {
          print "\\end{longtable}\n\\end{scriptsize}\n\\end{document}\n";
     }
}

sub decoupe {
     my $ligne = shift;
     my @champs;

     # CRLF, CR, LF
     $ligne =~ tr/\r\n//d;
     $ligne =~ s/\|$//;
     @champs = split(/\|/, $ligne);
     return \@champs;
}

sub lecture_table {
     my $chemin = shift;
     my ($ligne, $decalage, $entree);
     my %table;
     my $entete;
     my $info;
     my $display = $chemin;

     if ($chemin) {
          open (FICHIER, "<$chemin") || die "Can't open $display: $!\n";
     } else {
          *FICHIER = *STDIN;
          $display = "(STDIN)";
     }
     # Reading preamble
     $ligne = <FICHIER>;
     ($entete = &decoupe($ligne)) ||
          die "$display: can't read file header description.\n";
     $ligne = <FICHIER>;
     ($info = &decoupe($ligne)) ||
          die "$display: can't read file informations.\n";
     ($#$entete == $#$info) || die "$display: incoherent file header.\n";
     for $decalage (0..$#$entete) {
          $table{'origine'}{$entete->[$decalage]} = $info->[$decalage];
     }

     # Reading field names
     $ligne = <FICHIER>;
     # Remove optionnal headers : between [ and ]
     $ligne =~ s/\[[^\]]*\]//g;
     ($entete = &decoupe($ligne)) ||
          die "$display: can't read field names.\n";
     for $decalage (0..$#$entete) {
          $table{'champs'}{$entete->[$decalage]} = $decalage;
          print STDERR "Header analysis, field #$decalage = ".
               "'$entete->[$decalage]'\n" if $opt_D;
     }

     # Reading datas
     my $ref = \@{$table{'donnes'}};
     for ($entree = 0; $ligne = <FICHIER>; $entree++) {
          $ref->[$entree] = &decoupe($ligne);
          ($#$entete > $#{$ref->[$entree]}) &&
               ($#{$ref->[$entree]} = $#$entete);
     }
     print STDERR "$display: $entree data line(s) read.\n" if $opt_D;

     close(FICHIER) if $chemin;
     return \%table;
}

