/******************************************************************************
 *
 * dt_error.c - errors handling routine. Part of the DirTools project.
 *
 * $Revision: 1.2 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#ifdef _WIN32
#include <windows.h>
#else
#include "wincompat.h"
#endif /* _WIN32 */

#include "dt_error.h"

FILE *FErrorOutput = NULL;

/**********************************************************************
 * Alarm : print error messages to FErrorOutput
 *
 * - [in] taLevel : error level
 * - [in] pMessage : specific error message (format string)
 * - [in] : variable number of arguments according to previous format
 *          string
 *
 *********************************************************************/
void
Alarm (const TYPE_ALARM taLevel, const char *pMessage, ...)
{

     va_list args;

     if (!FErrorOutput)
          FErrorOutput = stderr;

     fflush(NULL);

     switch (taLevel) {
#ifdef _WIN32
     case TA_SYS_FATAL:
          fprintf (FErrorOutput,
                   "\nFATAL %s (%lu) - Aborting - ",
                   strerror (GetLastError ()), GetLastError ());
          break;
     case TA_SYS_ERR:
          fprintf (FErrorOutput,
                   "\nERROR %s (%lu) - ",
                   strerror (GetLastError ()), GetLastError ());
          break;
     case TA_SYS_WARN:
          fprintf (FErrorOutput,
                   "\nWARNING %s (%lu) - ",
                   strerror (GetLastError ()), GetLastError ());
          break;
#else /* _WIN32 */
     case TA_SYS_FATAL:
          fprintf (FErrorOutput,
                   "\nFATAL %s (%u) - Aborting - ", strerror (errno), errno);
          break;
     case TA_SYS_ERR:
          fprintf (FErrorOutput,
                   "\nERROR %s (%u) - ", strerror (errno), errno);
          break;
     case TA_SYS_WARN:
          fprintf (FErrorOutput,
                   "\nWARNING %s (%u) - ", strerror (errno), errno);
          break;
#endif /* _WIN32 */
     case TA_PROG_FATAL:
          fprintf (FErrorOutput, "\nFATAL - Aborting - ");
          break;
     case TA_PROG_ERR:
          fprintf (FErrorOutput, "\nERROR - ");
          break;
     case TA_PROG_WARN:
          fprintf (FErrorOutput, "\nWARNING - ");
          break;
     default:
          fprintf (FErrorOutput, "\nUNKNOWN ERROR %u - Aborting\n", taLevel);
          exit (EXIT_FAILURE);
          break;
     }

     if (pMessage == NULL) {
          fprintf (FErrorOutput, "UNDEFINED MESSAGE");
     } else {
          va_start (args, pMessage);
          vfprintf (FErrorOutput, pMessage, args);
          va_end (args);
     }

     fprintf (FErrorOutput, "\n\n");

     fflush (FErrorOutput);

     if (taLevel & TA_FATAL) {
          _fcloseall ();
          exit (EXIT_FAILURE);
     }
}
